"""Utilities for Flask. View classes, decorators, URL route converters, etc."""
import functools
import html
import logging
import os
import re
import urllib.parse

import flask
from flask import abort, get_flashed_messages, make_response, redirect, render_template, request, Response
from flask.views import View
from google.cloud import ndb
import requests
import werkzeug.exceptions
from werkzeug.exceptions import BadRequestKeyError, HTTPException
from werkzeug.routing import BaseConverter

from .appengine_info import LOCAL_SERVER
from . import util

logger = logging.getLogger(__name__)

# Modern HTTP headers for CORS, CSP, other security, etc.
CSP_HOSTS = 'localhost:8080 127.0.0.1:8080 my.dev.com:8080' if LOCAL_SERVER else ''
MODERN_HEADERS = {
  'Access-Control-Allow-Headers': '*',
  'Access-Control-Allow-Methods': '*',
  'Access-Control-Allow-Origin': '*',
  # https://developer.mozilla.org/en-US/docs/Web/HTTP/Reference/Headers/Cross-Origin-Opener-Policy
  'Cross-Origin-Opener-Policy': 'same-origin',
  # see https://content-security-policy.com/
  'Content-Security-Policy':
    f"script-src https: {CSP_HOSTS} 'unsafe-inline'; frame-ancestors 'self'",
  # 16070400 seconds is 6 months
  'Strict-Transport-Security': 'max-age=16070400; preload',
  'X-Content-Type-Options': 'nosniff',
  # https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/X-Frame-Options
  'X-Frame-Options': 'SAMEORIGIN',
  'X-XSS-Protection': '1; mode=block',
}

# https://cloud.google.com/tasks/docs/creating-appengine-handlers#reading-headers
CLOUD_TASKS_TASK_HEADER = 'X-AppEngine-TaskName'

# https://cloud.google.com/appengine/docs/standard/scheduling-jobs-with-cron-yaml#securing_urls_for_cron
APP_ENGINE_CRON_HEADER = 'X-Appengine-Cron'

# A few extra non-error HTTPExceptions
class Created(HTTPException):
    code = 201
    description = 'Created'

class Accepted(HTTPException):
    code = 202
    description = 'Accepted'

class NoContent(HTTPException):
    code = 204
    description = 'No Content'

class Redirect(HTTPException):
    def __init__(self, *args, location=None, headers=None, **kwargs):
      # this evidently isn't provided when flask-caching unpickles a pickled instance
      # assert location
      self.location = location
      self.headers = headers or {}
      super().__init__(**kwargs)

    def get_headers(self, *args, **kwargs):
      return {
        **self.headers,
        'Location': self.location,
      }

class MovedPermanently(Redirect):
    code = 301
    description = 'Moved Permanently'

class Found(Redirect):
    code = 302
    description = 'Found'

class NotModified(HTTPException):
    code = 304
    description = 'Not Modified'

class PaymentRequired(HTTPException):
    code = 402
    description = 'Payment Required'

class ProxyAuthenticationRequired(HTTPException):
    code = 407
    description = 'Proxy Authentication Required'

class MisdirectedRequest(HTTPException):
    code = 421
    description = 'Misdirected Request'

class UpgradeRequired(HTTPException):
    code = 426
    description = 'Upgrade Required'

class PreconditionRequired(HTTPException):
    code = 428
    description = 'Precondition Required'

class ClientClosedRequest(HTTPException):
    code = 499
    description = 'Client Closed Request'

class VariantAlsoNegotiates(HTTPException):
    code = 506
    description = 'Variant Also Negotiates'

class InsufficientStorage(HTTPException):
    code = 507
    description = 'Insufficient Storage'

class LoopDetected(HTTPException):
    code = 508
    description = 'Loop Detected'

class NotExtended(HTTPException):
    code = 510
    description = 'Not Extended'

class NetworkAuthenticationRequired(HTTPException):
    code = 511
    description = 'Network Authentication Required'

class NetworkConnectTimeoutError(HTTPException):
    code = 599
    description = 'Network Connect Timeout Error'


for cls in (
    Created,
    Accepted,
    NoContent,
    MovedPermanently,
    Found,
    NotModified,
    PaymentRequired,
    ProxyAuthenticationRequired,
    MisdirectedRequest,
    UpgradeRequired,
    PreconditionRequired,
    ClientClosedRequest,
    VariantAlsoNegotiates,
    InsufficientStorage,
    LoopDetected,
    NotExtended,
    NetworkAuthenticationRequired,
    NetworkConnectTimeoutError,
):
  # https://github.com/pallets/flask/issues/1837#issuecomment-304996942
  werkzeug.exceptions.default_exceptions.setdefault(cls.code, cls)
  werkzeug.exceptions._aborter.mapping.setdefault(cls.code, cls)


class RegexConverter(BaseConverter):
  """Regexp URL route for Werkzeug/Flask.

  Based on https://github.com/rhyselsmore/flask-reggie.

  Usage::

      @app.route('/<regex("abc|def"):letters>')

  Install with::

      app.url_map.converters['regex'] = RegexConverter
  """
  def __init__(self, url_map, *items):
    super(RegexConverter, self).__init__(url_map)
    self.regex = items[0]


def get_required_param(name):
  """Returns the given request parameter.

  If it's not in a query parameter or POST field, the current HTTP request
  aborts with status 400.
  """
  try:
    val = request.values.get(name)
  except (UnicodeDecodeError, UnicodeEncodeError) as e:
    error(f"Couldn't decode parameters as UTF-8: {e}")

  if not val:
    error(f'Missing required parameter: {name}')

  return val


def ndb_context_middleware(app, client=None, **kwargs):
  """WSGI middleware to add an NDB context per request.

  Follows the WSGI standard. Details: http://www.python.org/dev/peps/pep-0333/

  Install with eg::

      ndb_client = ndb.Client()
      app = Flask('my-app')
      app.wsgi_app = flask_util.ndb_context_middleware(app.wsgi_app, ndb_client)

  Background: https://cloud.google.com/appengine/docs/standard/python3/migrating-to-cloud-ndb#using_a_runtime_context_with_wsgi_frameworks

  Args:
    client: :class:`google.cloud.ndb.Client`
    kwargs: passed through to :meth:`google.cloud.ndb.Client.context`
  """
  def wrapper(environ, start_response):
    if ndb.context.get_context(raise_context_error=False):
      # someone else (eg a unit test harness) has already created a context
      return app(environ, start_response)

    with client.context(**kwargs):
      return app(environ, start_response)

  wrapper.kwargs = kwargs
  return wrapper


def handle_exception(e):
  """Flask error handler that propagates HTTP exceptions into the response.

  Install with::

      app.register_error_handler(Exception, handle_exception)
  """
  if isinstance(e, BadRequestKeyError):
      if e.args:
          e._description = f'Missing required parameter: {e.args[0]}'
      else:
          e.show_exception = True

  if isinstance(e, HTTPException):
    # raised by this app itself, pass it through. use body and headers from
    # response if available (but not status code).
    resp = e.get_response()
    if resp:
      resp.status_code = e.code
      return resp
    else:
      return str(e), e.code, e.get_headers()

  code, body = util.interpret_http_exception(e)
  if code:
    return ((f'Upstream server request failed: {e}' if code in ('502', '504')
             else f'HTTP Error {code}: {body}'),
            int(code),
            {'Content-Type': 'text/plain; charset=utf-8'})

  raise e


def error(msg, status=400, exc_info=False, **kwargs):
  """Logs and returns an HTTP error via :class:`werkzeug.exceptions.HTTPException`.

  Args:
    msg (str)
    status (int)
    exc_info: Python exception info three-tuple, eg from :func:`sys.exc_info`
    kwargs: passed through to :func:`flask.abort`
  """
  logger.info(f'Returning {status}: {msg} {kwargs}', exc_info=exc_info)
  try:
    abort(int(status), msg, **kwargs)
  except LookupError:  # probably an unknown status code
    raise HTTPException(response=Response(response=msg, status=status), **kwargs)


def flash(msg, escape=True, **kwargs):
  """Wrapper for :func:`flask.flash`` that also logs the message.

  Args:
    msg (str)
    escape (bool): whether to HTML-escape ``msg``
  """
  if escape:
    msg = html.escape(msg, quote=False)
  flask.flash(msg, **kwargs)
  logger.info(f'Flashed message: {msg}')


def default_modern_headers(resp):
  """Include modern HTTP headers by default, but let the response override them.

  Install with::

      app.after_request(default_modern_headers)
  """
  for name, value in MODERN_HEADERS.items():
    resp.headers.setdefault(name, value)

  return resp


def cached(cache, timeout, headers=(), http_5xx=False):
  """Thin flask-cache wrapper that supports timedelta and cache query param.

  If the ``cache`` URL query parameter is ``false``, skips the cache. Also, does
  not store the response in the cache if it's an HTTP 5xx or if there are any
  flashed messages.

  Args:
    cache (:class:`flask_caching.Cache`)
    timeout (:class:`datetime.timedelta`)
    headers: sequence of str, optional headers to include in the cache key
    http_5xx (bool): optional, whether to cache HTTP 5xx (server error) responses
  """
  # TODO: make new thread-safe Cache subclass
  # for eg https://console.cloud.google.com/errors/detail/CKL6udCe3IuR9QE;time=P30D?project=brid-gy
  def response_filter(resp):
    """Return False if the response shouldn't be cached."""
    resp = make_response(resp)
    return (not get_flashed_messages() and 'Set-Cookie' not in resp.headers and
            (http_5xx or resp.status_code // 100 != 5))

  def unless():
    return bool(request.args.get('cache', '').lower() == 'false' or
                request.cookies)

  def decorator(f):
    # catch werkzeug HTTPExceptions, eg raised by abort(), and return them
    # instead of letting them propagate, so that flask-cache can cache them
    @functools.wraps(f)
    def httpexception_to_return(*args, **kwargs):
      try:
        return f(*args, **kwargs)
      except HTTPException as e:
        return e

    decorated = cache.cached(
      timeout.total_seconds(),
      query_string=True,
      response_filter=response_filter,
      unless=unless,
    )(httpexception_to_return)

    # include specified headers in cache key:
    # https://flask-caching.readthedocs.io/en/latest/api.html#flask_caching.Cache.cached
    orig_cache_key = decorated.make_cache_key
    def make_cache_key(*args, **kwargs):
      header_vals = '  '.join(request.headers.get(h, '') for h in sorted(headers))
      # an alternative to including host_url would be to pass
      # key_prefix=f'view/{request.base_url}' to cache.cached above, but
      # flask-caching doesn't currently support query_string and key_prefix
      # together :(
      # https://github.com/pallets-eco/flask-caching/issues/302
      k = f'{request.host_url} {orig_cache_key(*args, **kwargs)}  {header_vals}'
      if request.method != 'GET':
        k = f'{request.method} {k}'
      return k

    decorated.make_cache_key = make_cache_key

    return decorated

  return decorator


def headers(headers, error_codes=(404,)):
  """Flask decorator that adds headers to the response.

  Args:
    headers (dict mapping str header name to str value)
    error_codes (sequence of int): 4xx and 5xx HTTP codes to include the headers
      with, along with 2xx and 3xx.
  """
  def decorator(fn):
    @functools.wraps(fn)
    def decorated(*args, **kwargs):
      try:
        ret = fn(*args, **kwargs)
      except HTTPException as e:
        if e.code in error_codes:
          if not e.response:
            e.response = make_response(html.escape(e.description), e.code, e.get_headers())
          e.response.headers.update(headers)
        raise

      resp = make_response(ret)
      resp.headers.update(headers)
      return resp

    return decorated

  return decorator


def cloud_tasks_only(log=True):
  """Flask decorator that returns HTTP 401 if the request isn't from Cloud Tasks.

  (...or from App Engine Cron.)

  Ignored if ``LOCAL_SERVER`` is True.

  https://cloud.google.com/tasks/docs/creating-appengine-handlers#reading-headers
  https://cloud.google.com/appengine/docs/standard/scheduling-jobs-with-cron-yaml#securing_urls_for_cron

  Must be used *below* :meth:`flask.Flask.route`, eg::

      @app.route('/path')
      @cloud_tasks_only()
      def handler():
          ...

  Args:
    log (boolean): whether to log the task name. If None, task name is logged
      only if the traceparent HTTP header is not set.
  """
  def decorator(fn):
    @functools.wraps(fn)
    def decorated(*args, **kwargs):
      task = request.headers.get('X-AppEngine-TaskName')
      if (not task and APP_ENGINE_CRON_HEADER not in request.headers
          and not LOCAL_SERVER):
        return 'Internal only', 401

      if log or (log is None and not request.headers.get('traceparent')):
        logger.info(f"Task {task}")

      return fn(*args, **kwargs)

    return decorated

  return decorator


def canonicalize_domain(from_domains, to_domain):
  """WSGI middleware that redirects one or more domains to a canonical domain.

  Preserves scheme, path, and query.

  Install with eg::

      app = flask.Flask(...)
      app.before_request(canonicalize_domain(('old1.com', 'old2.org'), 'new.com'))

  Args:
    from_domains: str or sequence of str
    to_domain: str
  """
  if isinstance(from_domains, str):
    from_domains = [from_domains]

  def fn():
    parts = list(urllib.parse.urlparse(request.url))
    # not using request.host because it includes port
    if parts[1] in from_domains:  # netloc
      parts[1] = to_domain
      return redirect(urllib.parse.urlunparse(parts), code=301)

  return fn


def canonicalize_request_domain(from_domains, to_domain):
  """Flask handler decorator that redirects to a canonical domain.


  Use *below* :meth:`flask.Flask.route`, eg::

      @app.route('/path')
      @canonicalize_request_domain('foo.com', 'bar.com')
      def handler():
          ...

  Args:
    from_domains: str or sequence of str
    to_domain: str
  """
  def decorator(fn):
    @functools.wraps(fn)
    def decorated(*args, **kwargs):
      return canonicalize_domain(from_domains, to_domain)() or fn(*args, **kwargs)

    return decorated

  return decorator


class XrdOrJrd(View):
  """Renders and serves an XRD or JRD file.

  JRD is served if the request path ends in .jrd or .json, or the format query
  parameter is ``jrd`` or ``json``, or the request`s Accept header includes
  ``jrd`` or ``json``.

  XRD is served if the request path ends in .xrd or .xml, or the format query
  parameter is ``xml`` or ``xrd``, or the request's Accept header includes
  ``xml`` or ``xrd``.

  Otherwise, defaults to DEFAULT_TYPE.

  Subclasses must override :meth:`template_prefix()`` and
  :meth:`template_vars()``. URL route variables are passed through to
  :meth:`template_vars()`` as keyword args.
  """
  JRD = 'jrd'
  XRD = 'xrd'
  DEFAULT_TYPE = JRD
  """Either ``JRD`` or ``which``, the type to return by default if the request
  doesn't ask for one explicitly with the Accept header."""

  def template_prefix(self):
    """Returns template filename, without extension."""
    raise NotImplementedError()

  def template_vars(self, **kwargs):
    """Returns a dict with template variables.

    URL route variables are passed through as kwargs.
    """
    raise NotImplementedError()

  def _type(self):
    """Returns XRD or JRD."""
    format = request.args.get('format', '').lower()
    ext = os.path.splitext(request.path)[1]

    if ext in ('.jrd', '.json') or format in ('jrd', 'json'):
      return self.JRD
    elif ext in ('.xrd', '.xml') or format in ('xrd', 'xml'):
      return self.XRD

    # We don't do full content negotiation (Accept Header parsing); we just
    # check whether jrd/json and xrd/xml are in the header, and if they both
    # are, which one comes first. :/
    # https://developer.mozilla.org/en-US/docs/Web/HTTP/Content_negotiation
    accept = request.headers.get('Accept', '').lower()
    jrd = re.search(r'jrd|json', accept)
    xrd = re.search(r'xrd|xml', accept)
    if jrd and (not xrd or jrd.start() < xrd.start()):
      return self.JRD
    elif xrd and (not jrd or xrd.start() < jrd.start()):
      return self.XRD

    assert self.DEFAULT_TYPE in (self.JRD, self.XRD)
    return self.DEFAULT_TYPE

  @headers({'Vary': 'Accept'})
  def dispatch_request(self, **kwargs):
    data = self.template_vars(**kwargs)
    assert isinstance(data, dict)

    # Content-Types are from https://tools.ietf.org/html/rfc7033#section-10.2
    if self._type() == self.JRD:
        return data, {'Content-Type': 'application/jrd+json'}

    template = f'{self.template_prefix()}.{self._type()}'
    return (render_template(template, **data),
            {'Content-Type': 'application/xrd+xml; charset=utf-8'})


class FlashErrors(View):
    """Wraps a Flask :class:`flask.view.View` and flashes errors.

    Mostly used with OAuth endpoints.
    """
    def dispatch_request(self):
        try:
            return super().dispatch_request()
        except (ValueError, requests.RequestException) as e:
            logger.warning(f'{self.__class__.__name__} error', exc_info=True)
            _, body = util.interpret_http_exception(e)
            flash(util.linkify(body or str(e), pretty=True))
            return redirect('/login')



"""Meetup.com drop-in.

API docs:
https://www.meetup.com/meetup_api/
"""
import logging
import urllib.parse

from flask import request
from google.cloud import ndb

from . import views
from .models import BaseAuth
from .webutil import appengine_info, flask_util, util
from .webutil.util import json_loads

logger = logging.getLogger(__name__)

if appengine_info.DEBUG:
  MEETUP_CLIENT_ID = util.read('meetup_client_id_local')
  MEETUP_CLIENT_SECRET = util.read('meetup_client_secret_local')
else:
  MEETUP_CLIENT_ID = util.read('meetup_client_id')
  MEETUP_CLIENT_SECRET = util.read('meetup_client_secret')

GET_AUTH_CODE_URL = '&'.join((
  'https://secure.meetup.com/oauth2/authorize?client_id=%(client_id)s',
  'response_type=code',
  'redirect_uri=%(redirect_uri)s',
  'scope=%(scope)s',
  'state=%(state)s',
))

GET_ACCESS_TOKEN_URL = 'https://secure.meetup.com/oauth2/access'
GET_USER_INFO_URL = 'https://api.meetup.com/members/self/'


def urlopen_bearer_token(url, access_token, data=None, **kwargs):
  """Wraps urlopen() and adds OAuth credentials to the request.
  """
  headers = {'Authorization': f'Bearer {access_token}'}
  try:
    return util.urlopen(urllib.request.Request(url, headers=headers, data=data), **kwargs)
  except BaseException as e:
    util.interpret_http_exception(e)
    raise


class MeetupAuth(BaseAuth):
  """An authenticated Meetup.com user.

  Provides methods that return information about this user and make
  OAuth-signed requests to Meetup's HTTP-based APIs. Stores OAuth credentials
  in the datastore. See models.BaseAuth for usage details.

  Implements urlopen() but not api().
  """
  access_token_str = ndb.StringProperty(required=True)
  user_json = ndb.TextProperty(required=True)

  def site_name(self):
    return 'Meetup.com'

  def user_display_name(self):
    """Returns the Meetup.com user id."""
    return json_loads(self.user_json)['name']

  def image_url(self):
    """Returns the user's profile picture URL, if any."""
    return json_loads(self.user_json).get('avatar')

  def access_token(self):
    """Returns the OAuth access token string."""
    return self.access_token_str

  def urlopen(self, url, **kwargs):
    return urlopen_bearer_token(url, self.access_token_str, kwargs)


class MeetupCsrf(ndb.Model):
  """Stores a CSRF token for the Meetup.com OAuth2 flow."""
  token = ndb.StringProperty(required=False)
  state = ndb.TextProperty(required=False)


class Start(views.Start):
  """Starts Meetup.com auth. Requests an auth code and expects a redirect back.
  """

  NAME = 'meetup'
  LABEL = 'Meetup.com'
  DEFAULT_SCOPE = ''

  def redirect_url(self, state=None):

    assert MEETUP_CLIENT_ID and MEETUP_CLIENT_SECRET, \
      "Please fill in the meetup_client_id and meetup_client_secret files in your app's root directory."

    csrf_key = MeetupCsrf(state=state).put()

    return GET_AUTH_CODE_URL % {
      'client_id': MEETUP_CLIENT_ID,
      'redirect_uri': urllib.parse.quote_plus(self.to_url()),
      'scope': self.scope,
      'state': f'{urllib.parse.quote_plus(state or "")}|{csrf_key.id()}',
    }

  @classmethod
  def button_html(cls, *args, **kwargs):
    return super(cls, cls).button_html(
      *args, input_style='background-color: #EEEEEE; padding: 10px', **kwargs)


class Callback(views.Callback):
  """The auth callback. Fetches an access token, stores it, and redirects home.
  """
  def dispatch_request(self):
    # handle errors
    error = request.values.get('error')
    if error:
      if error == 'access_denied':
        logger.info('User declined')
        return self.finish(None, state=request.values.get('state'))
      else:
        flask_util.error(f"Error: {error}: {request.values.get('error_description')}")

    state = request.values['state']
    # lookup the CSRF token
    try:
      csrf_id = int(urllib.parse.unquote_plus(state).split('|')[-1])
    except (ValueError, TypeError):
      flask_util.error(f'Invalid state value {state!r}')

    csrf = MeetupCsrf.get_by_id(csrf_id)
    if not csrf:
      flask_util.error(f'No CSRF token for id {csrf_id}')

    # extract auth code and request access token
    auth_code = request.values['code']
    data = {
      'code': auth_code,
      'client_id': MEETUP_CLIENT_ID,
      'client_secret': MEETUP_CLIENT_SECRET,
      'grant_type': 'authorization_code',
      # redirect_uri here must be the same in the oauth code request!
      # (the value here doesn't actually matter since it's requested server side.)
      'redirect_uri': request.base_url,
    }
    # TODO: handle refresh tokens
    resp = util.requests_post(GET_ACCESS_TOKEN_URL, data=data)
    resp.raise_for_status()

    logger.debug(f'Access token response: {resp}')

    try:
      data = json_loads(resp.text)
    except (ValueError, TypeError):
      logger.error(f'Bad response:\n{resp}', exc_info=True)
      flask_util.error('Bad Disqus response to access token request')

    error = data.get('error')
    if error:
      msg = f"Error: {error[0]}: {data.get('error_description')}"
      flask_util.error(msg)

    access_token = data['access_token']

    user_json = urlopen_bearer_token(GET_USER_INFO_URL, access_token).read()
    user = json_loads(user_json)
    logger.debug(f'User info response: {user}')
    user_id = str(user['id'])

    logger.info(f'Storing new Meetup account for ID: {user_id}')
    auth = MeetupAuth(id=user_id, access_token_str=access_token, user_json=user_json)
    auth.put()
    return self.finish(auth, state=csrf.state)



"""Unit tests for models.py.
"""
import enum
import warnings

from cryptography.hazmat.primitives.ciphers.aead import AESGCM
from google.cloud import ndb

import unittest.mock as mock

from ..models import EncryptedProperty, EnumProperty, StringIdModel
from .. import appengine_config, models, testutil


class StringIdModelTest(testutil.TestCase):

  def setUp(self):
    warnings.filterwarnings('ignore', module='google.auth',
      message='Your application has authenticated using end user credentials')

  def test_put(self):
    with appengine_config.ndb_client.context():
      self.assertEqual(ndb.Key(StringIdModel, 'x'),
                       StringIdModel(id='x').put())
      self.assertRaises(AssertionError, StringIdModel().put)
      self.assertRaises(AssertionError, StringIdModel(id=1).put)


class TestEnum(enum.Enum):
  FOO = 1
  BAR = 2


class EnumModel(ndb.Model):
  field = EnumProperty(TestEnum)


class EnumPropertyTest(testutil.TestCase):

  def test_init_non_enum(self):
    with self.assertRaises(TypeError):
      EnumProperty(str)

  def test_validate(self):
    with self.assertRaises(TypeError):
      EnumModel(field='not_an_enum').put()

  def test_round_trip(self):
    with appengine_config.ndb_client.context():
      entity = EnumModel()
      entity.put()
      self.assertIsNone(entity.key.get().field)

      entity.field = TestEnum.FOO
      entity.put()
      got = entity.key.get()
      self.assertEqual(TestEnum.FOO, got.field)
      self.assertEqual(1, got.field.value)

      entity.field = TestEnum.BAR
      entity.put()
      got = entity.key.get()
      self.assertEqual(TestEnum.BAR, got.field)
      self.assertEqual(2, got.field.value)


class EncryptedModel(ndb.Model):
  secret = EncryptedProperty()


class EncryptedPropertyTest(testutil.TestCase):
  def setUp(self):
    super().setUp()

    self.ndb_context = appengine_config.ndb_client.context()
    self.ndb_context.__enter__()

  def tearDown(self):
    self.ndb_context.__exit__(None, None, None)
    super().tearDown()

  def test_validate(self):
    with self.assertRaises(TypeError):
      EncryptedModel(secret=123).put()

    with self.assertRaises(TypeError):
      EncryptedModel(secret='string').put()

  def test_round_trip(self):
    entity = EncryptedModel(secret=b'seekret')
    entity.put()
    self.assertEqual(b'seekret', entity.key.get().secret)

    utf8 = 'émojis 🔐'.encode('utf-8')
    entity.secret = utf8
    entity.put()
    self.assertEqual(utf8, entity.key.get().secret)

  def test_encrypted_storage(self):
    test_secret = b'plaintext secret'
    entity = EncryptedModel(secret=test_secret)
    encrypted = EncryptedModel.secret._to_base_type(test_secret)

    self.assertIsInstance(encrypted, bytes)
    self.assertNotIn(test_secret, encrypted)
    self.assertEqual(12, len(encrypted[:12]))
    self.assertGreater(len(encrypted), 12 + len(test_secret))

    decrypted = EncryptedModel.secret._from_base_type(encrypted)
    self.assertEqual(test_secret, decrypted)

  def test_none_value(self):
    entity = EncryptedModel(secret=None)
    entity.put()
    self.assertIsNone(entity.key.get().secret)

  def test_no_key_error(self):
    self.mox.StubOutWithMock(models, 'ENCRYPTED_PROPERTY_KEY')
    models.ENCRYPTED_PROPERTY_KEY = None
    with self.assertRaises(RuntimeError) as cm:
      EncryptedModel(secret=b'test').put()
    self.assertIn('No encryption key found', str(cm.exception))

  def test_different_nonces(self):
    test_secret = b'same secret'
    prop = EncryptedModel.secret

    encrypted1 = prop._to_base_type(test_secret)
    encrypted2 = prop._to_base_type(test_secret)

    self.assertNotEqual(encrypted1, encrypted2)

    self.assertEqual(test_secret, prop._from_base_type(encrypted1))
    self.assertEqual(test_secret, prop._from_base_type(encrypted2))



"""Disqus OAuth drop-in.

Disqus API docs: https://disqus.com/api/docs/

This drop-in is even more similar to Instagram than Instagram is to
Facebook. Differences:

- urlopen must pass the ``api_key`` with each request (in addition to the
  ``access_token``)
- Response to access_token does not give much information about the user,
  so we additionally fetch ``/user/details`` before saving
- Deny appears to be broken on Disqus's side (clicking "No Thanks" has
  no effect), so we ignore that possibility for now.

TODO: unify Disqus, Facebook, and Instagram
"""
import logging
import urllib.parse

from flask import request
from google.cloud import ndb

from . import models, views
from .webutil import flask_util, util
from .webutil.util import json_dumps, json_loads

logger = logging.getLogger(__name__)

DISQUS_CLIENT_ID = util.read('disqus_client_id')
DISQUS_CLIENT_SECRET = util.read('disqus_client_secret')
GET_AUTH_CODE_URL = (
    'https://disqus.com/api/oauth/2.0/authorize/?' +
    '&'.join((
        'client_id=%(client_id)s',
        'scope=%(scope)s',
        'response_type=code',
        'redirect_uri=%(redirect_uri)s',
    )))
GET_ACCESS_TOKEN_URL = 'https://disqus.com/api/oauth/2.0/access_token/'
USER_DETAILS_URL = 'https://disqus.com/api/3.0/users/details.json?user=%d'


class DisqusAuth(models.BaseAuth):
  """An authenticated Disqus user.

  Provides methods that return information about this user (or page) and make
  OAuth-signed requests to Instagram's HTTP-based APIs. Stores OAuth credentials
  in the datastore. See :class:`models.BaseAuth` for usage details.

  Disqus-specific details: implements :meth:`urlopen` but not :meth:`api`.
  The key name is the Disqus user id.
  """
  auth_code = ndb.StringProperty(required=True)
  access_token_str = ndb.StringProperty(required=True)
  user_json = ndb.TextProperty(required=True)

  def site_name(self):
    return 'Disqus'

  def user_display_name(self):
    """Returns the user's name."""
    return json_loads(self.user_json)['name']

  def image_url(self):
    """Returns the user's profile picture URL, if any."""
    return json_loads(self.user_json).get('avatar', {}).get('permalink', {})

  def access_token(self):
    """Returns the OAuth access token string."""
    return self.access_token_str

  def urlopen(self, url, **kwargs):
    """Wraps :meth:`models.BaseAuth.urlopen` and adds OAuth credentials."""
    # TODO does work for POST requests? key is always passed as a
    # query param, regardless of method.
    return models.BaseAuth.urlopen_access_token(url, self.access_token_str,
                                                DISQUS_CLIENT_ID, **kwargs)


class Start(views.Start):
  """Starts Disqus auth. Requests an auth code and expects a redirect back.
  """
  NAME = 'disqus'
  LABEL = 'Disqus'

  # Disqus scopes are comma separated: read, write, admin, email
  # https://disqus.com/api/docs/requests/#data-availability
  DEFAULT_SCOPE = 'read'

  def redirect_url(self, state=None):
      assert DISQUS_CLIENT_ID and DISQUS_CLIENT_SECRET, \
          "Please fill in the disqus_client_id and disqus_client_secret files in your app's root directory."
      return GET_AUTH_CODE_URL % {
        'client_id': DISQUS_CLIENT_ID,
        'scope': self.scope,
        'redirect_uri': urllib.parse.quote_plus(self.to_url(state=state)),
      }


class Callback(views.Callback):
  """The auth callback. Fetches an access token, stores it, and redirects home.
  """
  def dispatch_request(self):
    if self.handle_error():
      return

    # https://disqus.com/api/docs/auth/
    auth_code = request.values['code']
    data = {
        'grant_type': 'authorization_code',
        'client_id': DISQUS_CLIENT_ID,
        'client_secret': DISQUS_CLIENT_SECRET,
        'redirect_uri': self.request_url_with_state(),
        'code': auth_code,
    }

    resp = util.requests_post(GET_ACCESS_TOKEN_URL, data=data)
    resp.raise_for_status()
    try:
      data = json_loads(resp.text)
    except (ValueError, TypeError):
      logger.error(f'Bad response:\n{resp}', exc_info=True)
      flask_util.error('Bad Disqus response to access token request')

    access_token = data['access_token']
    user_id = data['user_id']
    # TODO is a username key preferred?
    # username = data['username']

    auth = DisqusAuth(id=str(user_id),
                      auth_code=auth_code,
                      access_token_str=access_token)

    resp = auth.urlopen(USER_DETAILS_URL % user_id).read()
    try:
      user_data = json_loads(resp)['response']
    except (ValueError, TypeError):
      logger.error(f'Bad response:\n{resp}', exc_info=True)
      flask_util.error('Bad Disqus response to user details request')

    auth.user_json = json_dumps(user_data)
    logger.info(f'created disqus auth {auth}')
    auth.put()
    return self.finish(auth, state=request.values.get('state'))

  def handle_error(handler):
    """Handles any error reported in the callback query parameters.

    Args:
      handler (Callback)

    Returns:
      bool: True if there was an error, False otherwise
    """
    error = request.values.get('error')
    if error:
      if error == 'access_denied':
        logger.info('User declined')
        handler.finish(None, state=request.values.get('state'))
        return True
      else:
        flask_util.error(error)

    return False



"""HTTP request and server information derived from the environment.

App Engine:
* https://cloud.google.com/appengine/docs/standard/python3/runtime#environment_variables
* https://cloud.google.com/appengine/docs/flexible/python/runtime#environment_variables

WSGI: https://www.python.org/dev/peps/pep-3333/

To run against the prod datastore, make a service account, download its JSON
credentials file, then run eg:

env GOOGLE_APPLICATION_CREDENTIALS=service_account_creds.json FLASK_ENV=development flask run -p 8080
"""
import os, sys

PROJECT = os.getenv('GOOGLE_CLOUD_PROJECT') or os.getenv('GAE_APPLICATION') or ''
APP_ID = PROJECT.split('~')[-1]

creds = os.getenv('GOOGLE_APPLICATION_CREDENTIALS')
gae_env = os.environ.get('GAE_ENV')  # App Engine Standard
gae_instance = 'GAE_INSTANCE' in os.environ  # App Engine Flex

TESTING = 'unittest' in ' '.join(sys.argv)
if creds and not creds.endswith('fake_user_account.json'):
  DEBUG = False
  LOCAL_SERVER = True
else:
  DEBUG = gae_env in (None, 'localdev') and not gae_instance
  LOCAL_SERVER = gae_env != 'standard' and not gae_instance and not TESTING



"""Unit tests for flask_util.py."""
import datetime
import os
import unittest

from flask import abort, Flask, flash, get_flashed_messages, make_response, request
from flask_caching import Cache
from werkzeug.exceptions import BadRequest, HTTPException, NotFound

from .. import flask_util
from ..flask_util import get_required_param


class FlaskUtilTest(unittest.TestCase):
  def setUp(self):
    self.app = Flask('test_flask_util')
    self.app.url_map.converters['regex'] = flask_util.RegexConverter
    self.app.config.from_mapping({
      'TESTING': True,
      'SECRET_KEY': 'sooper seekret',
      'CACHE_TYPE': 'SimpleCache',
    })
    self.client = self.app.test_client()

  def test_regex_converter(self):
    @self.app.route('/<regex("abc|def"):letters>')
    def fn(letters):
      return ''

    with self.app.test_client() as client:
      resp = client.get('/def')
      self.assertEqual(200, resp.status_code)
      self.assertEqual('def', request.view_args['letters'])

      resp = client.get('/xyz')
      self.assertEqual(404, resp.status_code)

  def test_get_required_param(self):
    for ctx in (
        self.app.test_request_context('/?x=y&z='),
        self.app.test_request_context(method='POST', data={'x': 'y', 'z': ''}),
    ):
      with ctx:
        self.assertEqual('y', get_required_param('x'))
        with self.assertRaises(BadRequest):
            get_required_param('z')
        with self.assertRaises(BadRequest):
            get_required_param('a')

  def test_cached(self):
    cache = Cache(self.app)
    calls = 0

    def view():
      nonlocal calls
      calls += 1

      if 'flash' in request.args:
        flash('foo')

      resp = make_response(str(calls))

      if 'set-cookie' in request.args:
        resp.set_cookie('kooky')

      if '500' in request.args:
        resp.status_code = 500

      return resp

    @self.app.route('/foo')
    @flask_util.cached(cache, datetime.timedelta(days=1))
    def foo():
      return view()

    client = self.app.test_client(use_cookies=False)
    client.get('/foo?500')
    self.assertEqual(1, calls)

    client.get('/foo?500')
    self.assertEqual(2, calls)

    client.get('/foo?flash')
    self.assertEqual(3, calls)

    client.get('/foo?cache=false')
    self.assertEqual(4, calls)

    client.get('/foo?xyz')
    self.assertEqual(5, calls)

    client.get('/foo?abc')
    self.assertEqual(6, calls)

    resp = client.get('/foo?abc')
    self.assertEqual(6, calls)
    self.assertEqual('6', resp.get_data(as_text=True))

    client.get('/foo')
    self.assertEqual(7, calls)

    client.get('/foo', headers={'Cookie': 'bar'})
    self.assertEqual(8, calls)

    client.get('/foo?set-cookie')
    self.assertEqual(9, calls)

    resp = client.get('/foo')
    self.assertEqual(9, calls)
    self.assertEqual('7', resp.get_data(as_text=True))

    resp = client.head('/foo?head')
    self.assertEqual(10, calls)

    resp = client.get('/foo?head')
    self.assertEqual(11, calls)
    self.assertEqual('11', resp.get_data(as_text=True))

    # cache key should use full URL
    resp = client.get('/foo', base_url='https://foo/')
    self.assertEqual(12, calls)

    # cache key should include scheme
    resp = client.get('/foo', base_url='http://foo/')
    self.assertEqual(13, calls)

  def test_cached_http_5xx(self):
    cache = Cache(self.app)
    calls = 0

    @self.app.route('/error')
    @flask_util.cached(cache, datetime.timedelta(days=1), http_5xx=True)
    def error():
      nonlocal calls
      calls += 1
      return '', 500

    resp = self.client.get('/error')
    self.assertEqual(500, resp.status_code)
    self.assertEqual(1, calls)

    resp = self.client.get('/error')
    self.assertEqual(500, resp.status_code)
    self.assertEqual(1, calls)

  def test_cached_flask_util_error(self):
    cache = Cache(self.app)
    calls = 0

    @self.app.route('/error')
    @flask_util.cached(cache, datetime.timedelta(days=1))
    def error():
      nonlocal calls
      calls += 1
      flask_util.error('asdf', status=400)

    resp = self.client.get('/error')
    self.assertEqual(400, resp.status_code)
    self.assertEqual(1, calls)

    resp = self.client.get('/error')
    self.assertEqual(400, resp.status_code)
    self.assertEqual(1, calls)

  def test_error_unknown_status_code(self):
    with self.assertRaises(HTTPException) as e:
      flask_util.error('asdf', status=521)

    self.assertEqual(HTTPException, e.exception.__class__)
    self.assertEqual(521, e.exception.response.status_code)
    self.assertEqual('asdf', e.exception.response.get_data(as_text=True))

  def test_cached_headers(self):
    cache = Cache(self.app)
    calls = 0

    @self.app.route('/foo')
    @flask_util.cached(cache, datetime.timedelta(days=1),
                       headers=('Accept', 'Vary'))
    def foo():
      nonlocal calls
      calls += 1
      return ''

    client = self.app.test_client()
    client.get('/foo', headers={'Accept': 'bar'})
    self.assertEqual(1, calls)

    client.get('/foo', headers={'Accept': 'bar'})
    self.assertEqual(1, calls)

    client.get('/foo', headers={'Accept': 'baz'})
    self.assertEqual(2, calls)

    client.get('/foo', headers={'Accept': 'baz', 'Vary': 'biff'})
    self.assertEqual(3, calls)

    client.get('/foo', headers={'Accept': 'bar'})
    self.assertEqual(3, calls)

    client.get('/foo', headers={'Accept': 'baz', 'Vary': 'biff'})
    self.assertEqual(3, calls)

  def test_headers(self):
    ret = 'OK'
    headers = {'A': 'B', 'C': 'D'}

    @self.app.route('/foo')
    @flask_util.headers(headers)
    def foo():
      nonlocal ret
      return ret

    client = self.app.test_client()

    resp = client.get('/foo')
    self.assertEqual(200, resp.status_code)
    self.assertEqual('OK', resp.get_data(as_text=True))
    self.assertEqual('B', resp.headers['A'])
    self.assertEqual('D', resp.headers['C'])

    ret = ('?', {'X': 'Y'})
    resp = client.get('/foo')
    self.assertEqual(200, resp.status_code)
    self.assertEqual('?', resp.get_data(as_text=True))
    self.assertEqual('B', resp.headers['A'])
    self.assertEqual('D', resp.headers['C'])
    self.assertEqual('Y', resp.headers['X'])

    with self.app.test_request_context():
      ret = make_response(ret)

    resp = client.get('/foo')
    self.assertEqual(200, resp.status_code)
    self.assertEqual('?', resp.get_data(as_text=True))
    self.assertEqual('B', resp.headers['A'])
    self.assertEqual('D', resp.headers['C'])
    self.assertEqual('Y', resp.headers['X'])

  def test_headers_escape_exception_description(self):
    @self.app.route('/foo')
    @flask_util.headers({})
    def foo():
      raise NotFound('a<b')

    resp = self.app.test_client().get('/foo')
    self.assertEqual(404, resp.status_code)
    self.assertEqual('a&lt;b', resp.get_data(as_text=True))

  def test_headers_exception(self):
    @self.app.route('/foo')
    @flask_util.headers({'A': 'B'}, error_codes=(431,))
    def foo():
      abort(int(request.args['code']), 'fooey')

    client = self.app.test_client()

    resp = client.get('/foo?code=431')
    self.assertEqual(431, resp.status_code)
    self.assertEqual('B', resp.headers['A'])

    resp = client.get('/foo?code=501')
    self.assertEqual(501, resp.status_code)
    self.assertNotIn('A', resp.headers)

  def test_canonicalize_domain_get(self):
    @self.app.route('/', defaults={'_': ''})
    @self.app.route('/<path:_>')
    def view(_):
      return '', 204

    self.app.before_request(flask_util.canonicalize_domain('from.com', 'to.org'))

    # should redirect
    for url in '/', '/a/b/c', '/d?x=y':
      for scheme in 'http', 'https':
        resp = self.client.get(url, base_url=f'{scheme}://from.com')
        self.assertEqual(301, resp.status_code)
        self.assertEqual(f'{scheme}://to.org{url}', resp.headers['Location'])

    # shouldn't redirect
    for base_url in 'http://abc.net', 'https://to.org':
      resp = self.client.get('/', base_url=base_url)
      self.assertEqual(204, resp.status_code)
      self.assertNotIn('Location', resp.headers)

  def test_canonicalize_domain_post(self):
    @self.app.route('/<path:_>', methods=['POST'])
    def view(_):
      return '', 204

    self.app.before_request(
      flask_util.canonicalize_domain(('from.com', 'from.net'), 'to.org'))

    # should redirect and include *args
    for base_url in 'http://from.com', 'http://from.net':
      resp = self.client.post('/x/y', base_url=base_url)
      self.assertEqual(301, resp.status_code)
      self.assertEqual('http://to.org/x/y', resp.headers['Location'])

    # shouldn't redirect, should include *args
    for base_url in 'http://abc.net', 'https://to.org':
      resp = self.client.post('/x/y', base_url=base_url)
      self.assertEqual(204, resp.status_code)
      self.assertNotIn('Location', resp.headers)

  def test_canonicalize_request_domain_decorator(self):
    @self.app.route('/x')
    @flask_util.canonicalize_request_domain('foo.com', 'bar.com')
    def view():
      return 'OK'

    resp = self.client.get('/x')
    self.assertEqual(200, resp.status_code)

    resp = self.client.get('/x', base_url='http://foo.com/')
    self.assertEqual(301, resp.status_code)
    self.assertEqual('http://bar.com/x', resp.headers['Location'])

  def test_flash(self):
    with self.app.test_request_context('/'):
      flask_util.flash('foo')
      flask_util.flash('b<a>r')
      flask_util.flash('b<a>z', escape=False)
      self.assertEqual(['foo', 'b&lt;a&gt;r', 'b<a>z'], get_flashed_messages())

  def test_cloud_tasks_only(self):
    @self.app.route('/', methods=['POST'])
    @flask_util.cloud_tasks_only()
    def handler():
      return 'OK'

    got = self.client.post('/', headers={
        flask_util.CLOUD_TASKS_TASK_HEADER: '123',
    })
    self.assertEqual(200, got.status_code)
    self.assertEqual('OK', got.get_data(as_text=True))

    got = self.client.post('/')
    self.assertEqual(401, got.status_code)


class XrdOrJrdTest(unittest.TestCase):
  def setUp(self):
    super().setUp()

    class View(flask_util.XrdOrJrd):
      def template_prefix(self):
        return 'test_handler_template'

      def template_vars(self, **kwargs):
        return {'foo': 'bar'}

    self.View = View

    self.app = Flask('XrdOrJrdTest')
    self.app.config['TESTING'] = True
    self.app.template_folder = os.path.dirname(__file__)

    view_func = View.as_view('XrdOrJrdTest')
    self.app.add_url_rule('/', view_func=view_func)
    self.app.add_url_rule('/<path>', view_func=view_func)

    self.client = self.app.test_client()

  def assert_jrd(self, resp, expected={'foo': 'bar'}):
    self.assertEqual(200, resp.status_code)
    self.assertEqual('application/jrd+json', resp.headers['Content-Type'])
    self.assertEqual('Accept', resp.headers['Vary'])
    self.assertEqual(expected, resp.json)

  def assert_xrd(self, resp, expected='<XRD><Foo>bar</Foo></XRD>'):
    self.assertEqual(200, resp.status_code)
    self.assertEqual('application/xrd+xml; charset=utf-8',
                     resp.headers['Content-Type'])
    self.assertEqual('Accept', resp.headers['Vary'])
    self.assertEqual(expected, resp.get_data(as_text=True))

  def test_xrd_or_jrd_handler_default_jrd(self):
    self.assert_jrd(self.client.get('/'))
    for resp in (self.client.get('/x.xrd'),
                 self.client.get('/x.xml'),
                 self.client.get('/?format=xrd'),
                 self.client.get('/?format=xml'),
                 self.client.get('/', headers={'Accept': 'application/xrd+xml'}),
                 self.client.get('/', headers={'Accept': 'application/xml'}),
                 ):
      self.assert_xrd(resp)

  def test_xrd_or_jrd_handler_default_xrd(self):
    self.View.DEFAULT_TYPE = flask_util.XrdOrJrd.XRD

    self.assert_xrd(self.client.get('/'))
    for resp in (self.client.get('/x.jrd'),
                 self.client.get('/x.json'),
                 self.client.get('/?format=jrd'),
                 self.client.get('/?format=json'),
                 self.client.get('/', headers={'Accept': 'application/jrd+json'}),
                 self.client.get('/', headers={'Accept': 'application/json'}),
                 ):
      self.assert_jrd(resp)

  def test_xrd_or_jrd_handler_accept_header_order(self):
    self.assert_jrd(self.client.get('/', headers={
      'Accept': 'application/jrd+json,application/xrd+xml',
    }))
    self.assert_xrd(self.client.get('/', headers={
      'Accept': 'application/xrd+xml,application/jrd+json',
    }))



"""Bluesky auth drop-in. Supports both app password login and OAuth.

Use :class:`PasswordStart` and :class:`PasswordCallback` for app password,
class:`OAuthStart` and :class:`OAuthCallback` for OAuth.

https://atproto.com/specs/xrpc#:~:text=App,passwords
https://docs.bsky.app/docs/advanced-guides/oauth-client
https://atproto.com/specs/oauth
https://guillp.github.io/requests_oauth2client/
https://github.com/guillp/requests_oauth2client?tab=readme-ov-file#using-dpop
"""
import logging
import os
import re
from urllib.parse import quote, urljoin, urlparse

import arroba.did
from flask import redirect, request
from google.cloud import ndb
from lexrpc import Client
import requests
from requests_oauth2client import (
  AuthorizationRequestSerializer,
  DPoPToken,
  OAuth2Client,
  OAuth2Error,
  OAuth2AccessTokenAuth,
  TokenSerializer,
)

from . import views, models
from .webutil import flask_util, util
from .webutil.models import JsonProperty
from .webutil.util import json_dumps, json_loads

logger = logging.getLogger(__name__)

os.environ.setdefault('PLC_HOST', 'plc.directory')

# https://docs.bsky.app/docs/advanced-guides/oauth-client#client-and-server-metadata
PROTECTED_RESOURCE_PATH = '/.well-known/oauth-protected-resource'
RESOURCE_METADATA_PATH = '/.well-known/oauth-authorization-server'
CLIENT_METADATA_TEMPLATE = {
  # Clients must fill these in
  'client_id': None,      # eg 'https://app.example.com/oauth/client-metadata.json'
  'client_name': None,    # eg 'My Example App'
  'client_uri': None,     # eg 'https://app.example.com'
  'redirect_uris': None,  # eg ['https://app.example.com/oauth/callback'],

  # standard
  'application_type': 'web',
  'dpop_bound_access_tokens': True,
  'grant_types': [
    'authorization_code',
    'refresh_token',
  ],
  'response_types': ['code'],
  'scope': 'atproto transition:generic',
  'token_endpoint_auth_method': 'none',
}

_APP_CLIENT_METADATA = {
  **CLIENT_METADATA_TEMPLATE,
  'client_id': 'https://oauth-dropins.appspot.com/bluesky/client-metadata.json',
  'client_name': 'oauth-dropins demo',
  'client_uri': 'https://oauth-dropins.appspot.com/',
  'redirect_uris': ['https://oauth-dropins.appspot.com/bluesky/oauth_callback'],
}


def error(msg):
  logger.warning(msg)
  raise ValueError(msg)


class BlueskyLogin(ndb.Model):
  """An in-progress Bluesky OAuth login. Ephemeral.

  Stores a serialized :class:`requests_oauth2client.AuthorizationRequest` across
  HTTP requests.
  """
  state = ndb.TextProperty()
  did = ndb.StringProperty(required=True)
  authz_request = ndb.TextProperty(required=True)
  """Serialized :class:`requests_oauth2client.AuthorizationRequest`.

  Uses :meth:`requests_oauth2client.AuthorizationRequestSerializer.default_dumper` /
  :meth:`requests_oauth2client.AuthorizationRequestSerializer.default_loader`.
  """

  @classmethod
  def load(cls, id):
    if not util.is_int(id):
      error(f'State {id} not found')

    login = cls.get_by_id(int(id))
    if not login:
      error(f'State {id} not found')

    return login


class BlueskyAuth(models.BaseAuth):
  """An authenticated Bluesky user.

  Key id is DID.
  """
  password = ndb.StringProperty()
  pds_url = ndb.StringProperty()
  user_json = ndb.TextProperty(required=True)
  """app.bsky.actor.defs#profileViewDetailed"""
  session = JsonProperty()
  dpop_token = ndb.TextProperty()

  def site_name(self):
    return 'Bluesky'

  def access_token(self):
    """
    Returns:
      str:
    """
    if self.session:
      return self.session.get('accessJwt')

  def user_display_name(self):
    """
    Returns:
      str:
    """
    return json_loads(self.user_json).get('handle')

  def image_url(self):
    """
    Returns:
      str:
    """
    return json_loads(self.user_json).get('avatar')

  def oauth_api(self, client_metadata):
    """Returns an OAuth-based :class:`lexrpc.Client` for this user.

    Requires :attr:`dpop_token` to be set.

    Args:
      client_metadata (dict): client info metadata,
        https://docs.bsky.app/docs/advanced-guides/oauth-client#client-and-server-metadata

    Returns:
      lexrpc.Client:
    """
    assert self.dpop_token
    pds_url = self.pds_url or pds_for_did(self.key.id())
    oauth_client = oauth_client_for_pds(client_metadata, pds_url)
    dpop_token = TokenSerializer().loads(self.dpop_token)
    auth = OAuth2AccessTokenAuth(client=oauth_client, token=dpop_token)
    return Client(pds_url, auth=auth)

  def _api(self, **kwargs):
    """
    Args:
      kwargs: passed to the :class:`lexrpc.Client` constructor

    Returns:
      lexrpc.Client:
    """
    client = Client(address=self.pds_url, headers={'User-Agent': util.user_agent},
                    **kwargs)
    client.session = self.session
    return client

  @staticmethod
  def _api_from_password(handle, password, **kwargs):
    """
    Args:
      handle (str)
      password (str)
      kwargs: passed to the :class:`lexrpc.Client` constructor

    Returns:
      lexrpc.Client:
    """
    did = arroba.did.resolve_handle(handle)
    if not did:
      error(f"Couldn't resolve {handle} as a Bluesky handle")
    logger.info(f'resolved {handle} to {did}')
    pds_url = pds_for_did(did)

    logger.info(f'Logging into {pds_url} as {did}...')
    client = Client(address=pds_url, headers={'User-Agent': util.user_agent}, **kwargs)
    resp = client.com.atproto.server.createSession({
      'identifier': handle,
      'password': password,
    })
    return client


class StartBase(views.Start):
  """Base class for starting Bluesky auth; only used to provide the button.
  """
  NAME = 'bluesky'
  LABEL = 'Bluesky'

  @classmethod
  def button_html(cls, *args, **kwargs):
    kwargs['form_extra'] = kwargs.get('form_extra', '') + f"""
<input name="handle" type="text" class="form-control" placeholder="{cls.LABEL} handle" required style="width: 135px; height: 50px; display:inline;" />"""
    return super().button_html(
      *args,
      image_file='bluesky_logo.png',
      input_style='background-color: #EEEEEE',
      **kwargs)

Start = StartBase


class PasswordCallback(views.Callback):
  """
  App password login callback stub.
  """
  def dispatch_request(self):
    handle = util.remove_invisible_chars(request.values['username'])\
                 .strip().lower().removeprefix('@')
    password = request.values['password'].strip()
    state = request.values.get('state')

    # get the DID (portable user ID)
    try:
      client = BlueskyAuth._api_from_password(handle, password)
    except ValueError as e:
      logger.warning(f'Login failed: {e}')
      return self.finish(None, state=state)

    profile = {
      '$type': 'app.bsky.actor.defs#profileViewDetailed',
      **client.app.bsky.actor.getProfile(actor=handle)
    }
    auth = BlueskyAuth(
      id=profile['did'],
      password=password,
      # TODO: resolve DID's PDS
      pds_url = 'https://bsky.social',
      user_json=util.json_dumps(profile),
      session=client.session,
    )
    auth.put()
    return self.finish(auth, state=state)

Callback = PasswordCallback


def pds_for_did(did):
  """Resolves a DID document and extracts its PDS URL.

  https://atproto.com/specs/did#did-documents

  Args:
    did (str)

  Returns:
    str: PDS URL

  Raises:
    ValueError: if the DID couldn't be resolved, or if its DID document has no
    ATProto PDS endpoint
  """
  did_doc = arroba.did.resolve(did)
  if not did_doc:
    error(f"Couldn't resolve DID {did}")

  # based on bridgy_fed.atproto.ATProto.pds_for
  for service in did_doc.get('service', []):
    if service.get('id') in ('#atproto_pds', f'{did}#atproto_pds'):
      pds = service.get('serviceEndpoint')
      logger.info(f'{did} has PDS {pds}')
      return pds

  error(f"{did}'s DID doc has no ATProto PDS")


def oauth_client_for_pds(client_metadata, pds_url, redirect_uri=None):
  """Discovers a PDS's OAuth endpoints and creates a client.

  Args:
    client_metadata (dict)
    pds_url (str)
    redirect_uri (str): if not provided, defaults to the first element in
      ``redirect_uris`` in ``client_metadata`

  Returns:
    OAuth2Client:

  Raises:
    ValueError: if the DID couldn't be resolved, or if its DID document has no
    ATProto PDS endpoint
  """
  resp = util.requests_get(urljoin(pds_url, PROTECTED_RESOURCE_PATH))
  resp.raise_for_status()
  auth_server = resp.json()['authorization_servers'][0]
  logger.info(f'PDS {pds_url} has auth server {auth_server}')

  # OAuth special case for localhost client_id and redirect_uri
  # https://atproto.com/specs/oauth#:~:text=Localhost%20Client%20Development
  client_id = client_metadata['client_id']
  if not redirect_uri:
    redirect_uri = client_metadata['redirect_uris'][0]

  if request.host.split(':')[0] in ('localhost', '127.0.0.1'):
    redirect_uri = urljoin('http://127.0.0.1:8080', urlparse(redirect_uri).path)
    scope = quote(CLIENT_METADATA_TEMPLATE['scope'])
    client_id = f'http://localhost?redirect_uri={redirect_uri}&scope={scope}'

  return OAuth2Client.from_discovery_endpoint(
    urljoin(auth_server, RESOURCE_METADATA_PATH),
    client_id=client_id,
    redirect_uri=redirect_uri,
    dpop_bound_access_tokens=True,
  )


class OAuthStart(StartBase):
  """Starts the OAuth flow.

  Subclasses must populate:
    * :attr:`CLIENT_METADATA` (dict): client info metadata,
      https://docs.bsky.app/docs/advanced-guides/oauth-client#client-and-server-metadata
  """
  CLIENT_METADATA = None
  SCOPE = CLIENT_METADATA_TEMPLATE['scope']

  # available scopes as of Feb 2025:
  # atproto, transition:generic, transition:chat.bsky
  # https://bsky.social/.well-known/oauth-authorization-server
  DEFAULT_SCOPE = 'atproto transition:generic'
  SCOPE_SEPARATOR = ' '

  def redirect_url(self, state=None, handle=None):
    """Returns the URL for Bluesky to redirect back to after the OAuth prompt.

    Args:
      state (str): user-provided value to be returned as a query parameter in
        the return redirect
      handle (str): Bluesky domain handle. If ````None``, uses the ``handle``
        parameter in POST form data.

    Raises:
      ValueError, RequestException: if handle isn't a valid domain
    """
    assert self.CLIENT_METADATA
    client_id = self.CLIENT_METADATA['client_id']

    if not handle:
      handle = request.form['handle']

    handle = util.remove_invisible_chars(handle).strip().lstrip('@')
    if not re.fullmatch(util.DOMAIN_RE, handle):
      error(f"{handle} doesn't look like a domain")

    # resolve handle to DID doc and PDS base URL
    # https://atproto.com/specs/handle#handle-resolution
    did = arroba.did.resolve_handle(handle)
    if not did:
      error(f"Couldn't resolve {handle} as a Bluesky handle")
    logger.info(f'resolved {handle} to {did}')

    # generate authz URL, store session, redirect
    redirect_uri = self.to_url()
    if request.host.split(':')[0] in ('localhost', '127.0.0.1'):
      redirect_uri = urljoin('http://127.0.0.1:8080', urlparse(redirect_uri).path)

    client = oauth_client_for_pds(self.CLIENT_METADATA, pds_for_did(did),
                                  redirect_uri=redirect_uri)
    login_key = BlueskyLogin.allocate_ids(1)[0]

    try:
      authz_request = client.authorization_request(
        redirect_uri=redirect_uri, scope=self.SCOPE, state=login_key.id())
      par_request = client.pushed_authorization_request(authz_request)
    except OAuth2Error as e:
      error(e)

    serialized = AuthorizationRequestSerializer().dumps(authz_request)
    BlueskyLogin(key=login_key, state=state, did=did, authz_request=serialized).put()

    return par_request.uri


class OAuthCallback(views.Callback):
  """Finishes the OAuth flow.

  Subclasses must populate:
    * :attr:`CLIENT_METADATA` (dict): client info metadata,
      https://docs.bsky.app/docs/advanced-guides/oauth-client#client-and-server-metadata
  """
  CLIENT_METADATA = None

  def dispatch_request(self):
    # handle errors
    err = request.values.get('error')
    desc = request.values.get('error_description')
    if err:
      msg = f'Error: {err}: {desc}'
      logger.info(msg)
      if err == 'access_denied':
        return self.finish(None, state=request.values.get('state'))
      else:
        error(msg)

    login = BlueskyLogin.load(request.values['state'])
    pds_url = pds_for_did(login.did)
    client = oauth_client_for_pds(self.CLIENT_METADATA, pds_url,
                                  redirect_uri=request.url)

    # validate authz response, get access token
    try:
      authz_request = AuthorizationRequestSerializer().loads(login.authz_request)
      authz_resp = authz_request.validate_callback(request.url)
      token = client.authorization_code(authz_resp, validate=True)
    except OAuth2Error as e:
      error(e)

    if token.sub != login.did:
      error(f'Started login with {login.did} but authenticated {token.sub}')

    # get user profile
    # https://docs.bsky.app/docs/advanced-guides/oauth-client#callback-and-access-token-request
    auth = OAuth2AccessTokenAuth(client=client, token=token)
    pds_client = Client(pds_url, auth=auth)
    try:
      profile = pds_client.app.bsky.actor.getProfile(actor=login.did)
    except BaseException as e:
      code, body = util.interpret_http_exception(e)
      if code:
        error(f'{code} {body}')
      raise

    profile['$type'] = 'app.bsky.actor.defs#profileViewDetailed'
    auth = BlueskyAuth(id=login.did,
                       pds_url=pds_url,
                       dpop_token=TokenSerializer().dumps(token),
                       user_json=util.json_dumps(profile))
    auth.put()
    return self.finish(auth, state=login.state)



"""Threads OAuth 2 drop-in.

https://developers.facebook.com/docs/threads/
"""
import logging

from flask import request
from google.cloud import ndb
from requests.auth import HTTPBasicAuth
from requests_oauthlib import OAuth2Session

from . import models, views
from .webutil import flask_util, util
from .webutil.util import json_dumps, json_loads

logger = logging.getLogger(__name__)

APP_ID = util.read('threads_app_id')
APP_SECRET = util.read('threads_app_secret')

AUTH_CODE_URL = 'https://threads.net/oauth/authorize'
ACCESS_TOKEN_URL = 'https://graph.threads.net/oauth/access_token'
API_ACCOUNT_URL = 'https://graph.threads.net/v1.0/me?fields=id,username,name,threads_profile_picture_url,threads_biography'

# https://developers.facebook.com/docs/threads/get-started/get-access-tokens-and-permissions
ALL_SCOPES = (
  'threads_basic',
  'threads_content_publish',
  'threads_read_replies',
  'threads_manage_replies',
  'threads_manage_insights',
)


def https_if_localhost(url):
  return (url.replace('http://', 'https://', 1)
          if url.startswith('http://localhost:8080/')
          else url)


class ThreadsAuth(models.BaseAuth):
  """An OAuth-authenticated Threads user.

  Provides methods that return information about this user and store OAuth 2 tokens
  in the datastore. See models.BaseAuth for usage details.

  The datastore entity key name is the integer user id.
  """
  # Fields: token_type, access_token, scope, expires_at, expires_in
  token_json = ndb.TextProperty(required=True)
  user_json = ndb.TextProperty(required=True)

  def site_name(self):
    return 'Threads'

  def user_display_name(self):
    """Returns the username."""
    return json_loads(self.user_json).get('username')

  def image_url(self):
    """Returns the user's profile picture URL, if any."""
    return json_loads(self.user_json).get('threads_profile_picture_url')

  def access_token(self):
    """Returns the OAuth access token JSON."""
    return json_loads(self.token_json)['access_token']

  def session(self):
    """Returns a :class:`requests_oauthlib.OAuth2Session`."""
    token = json_loads(self.token_json)

    kwargs = {}
    if token.get('refresh_token') and token.get('expires_at'):
      def update_token(token):
        logging.info(f'Storing new access token {token}')
        self.token = token
        self.put()

      kwargs = {
        'auto_refresh_url': ACCESS_TOKEN_URL,
        'auto_refresh_kwargs': {'client_id': APP_ID},
        'token_updater': update_token,
      }

    session = OAuth2Session(APP_ID, token=token, **kwargs)
    session.auth = HTTPBasicAuth(APP_ID, APP_SECRET)
    return session


class Start(views.Start):
  """Starts three-legged OAuth with Threads.

  Redirects to Threads's auth prompt for user approval.
  """
  NAME = 'threads'
  LABEL = 'Threads'
  DEFAULT_SCOPE = 'threads_basic'

  @classmethod
  def button_html(cls, *args, **kwargs):
    return super(cls, cls).button_html(
      *args,
      input_style='background-color: #EEEEEE',
      **kwargs)

  def to_url(self, state=None):
    return https_if_localhost(super().to_url(state=state))

  def redirect_url(self, state=None):
    assert APP_ID and APP_SECRET, \
      "Please fill in the threads_app_key and threads_app_secret files in your app's root directory."

    # redirect to Threads auth URL
    session = OAuth2Session(APP_ID, scope=self.scope, redirect_uri=self.to_url())
    auth_url, state = session.authorization_url(AUTH_CODE_URL, state=state)
    logger.info(f'Redirecting to {auth_url}')
    return auth_url


class Callback(views.Callback):
  """The OAuth callback. Fetches an access token and redirects to the front page.
  """
  def dispatch_request(self):
    state = request.values.get('state')
    error = request.values.get('error')
    desc = request.values.get('error_description')
    if error:
      msg = f'Error: {error}: {desc}'
      logger.info(msg)
      if error == 'access_denied':
        return self.finish(None, state=state)
      else:
        flask_util.error(msg)

    session = OAuth2Session(APP_ID, redirect_uri=https_if_localhost(request.base_url))
    session.fetch_token(ACCESS_TOKEN_URL, include_client_id=True,
                        client_secret=APP_SECRET, authorization_response=request.url)
    logging.info(f'Got access token {session.token}')

    # Fetch user info
    resp = util.requests_get(API_ACCOUNT_URL, session=session)
    resp.raise_for_status()
    user_json = resp.json()
    logging.info(user_json)

    auth = ThreadsAuth(id=str(session.token['user_id']),
                       token_json=json_dumps(session.token),
                       user_json=json_dumps(user_json))
    auth.put()

    return self.finish(auth, state=request.values.get('state'))



"""App Engine datastore model base classes, properties, and utilites."""
import base64
from datetime import timezone
import enum
import os
from google.cloud import ndb

from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.ciphers.aead import AESGCM
from google.cloud.ndb.model import _BaseValue
from oauth_dropins.webutil import util
from oauth_dropins.webutil.util import json_dumps, json_loads

# 1MB limit: https://cloud.google.com/datastore/docs/concepts/limits
# use this to check an entity's size:
#   len(entity._to_pb().Encode())
MAX_ENTITY_SIZE = 1 * 1000 * 1000

ENCRYPTED_PROPERTY_KEY = ENCRYPTED_PROPERTY_KEY_BYTES = None
if key_base64 := util.read('encrypted_property_key'):  # base-64 encoded key bytes
  ENCRYPTED_PROPERTY_KEY_BYTES = base64.b64decode(key_base64)
  assert len(ENCRYPTED_PROPERTY_KEY_BYTES) == 32
  ENCRYPTED_PROPERTY_KEY = AESGCM(ENCRYPTED_PROPERTY_KEY_BYTES)


class StringIdModel(ndb.Model):
  """An :class:`ndb.Model` class that requires a string id."""
  def put(self, *args, **kwargs):
    """Raises AssertionError if string id is not provided."""
    assert self.key and self.key.string_id(), 'string id required but not provided'
    return super(StringIdModel, self).put(*args, **kwargs)


class JsonProperty(ndb.TextProperty):
    """Fork of ndb's that subclasses :class:`ndb.TextProperty` instead of :class:`ndb.BlobProperty`.

    This makes values show up as normal, human-readable, serialized JSON in the
    web console.
    https://github.com/googleapis/python-ndb/issues/874#issuecomment-1442753255

    Duplicated in arroba:
    https://github.com/snarfed/arroba/blob/main/arroba/ndb_storage.py
    """
    def _validate(self, value):
        if not isinstance(value, dict):
            raise TypeError('JSON property must be a dict')

    def _to_base_type(self, value):
        as_str = json_dumps(value, separators=(',', ':'), ensure_ascii=True)
        return as_str.encode('ascii')

    def _from_base_type(self, value):
        if not isinstance(value, str):
            value = value.decode('ascii')
        return json_loads(value)


class ComputedJsonProperty(JsonProperty, ndb.ComputedProperty):
    """Custom :class:`ndb.ComputedProperty` for JSON values that stores them as
    strings.

    ...instead of like :class:`ndb.StructuredProperty`, with "entity" type,
    which bloats them unnecessarily in the datastore.
    """
    def __init__(self, *args, **kwargs):
        kwargs['indexed'] = False
        super().__init__(*args, **kwargs)


class EnumProperty(ndb.IntegerProperty):
    """Property for storing Python Enum values.

    Stores the enum's integer value in the datastore.
    """
    def __init__(self, enum_class, **kwargs):
        if not issubclass(enum_class, enum.Enum):
            raise TypeError('enum_class must be a subclass of enum.Enum')
        self._enum_class = enum_class
        super().__init__(**kwargs)

    def _validate(self, value):
        if value is not None and not isinstance(value, self._enum_class):
            raise TypeError(f'Expected {self._enum_class.__name__}, got {type(value).__name__}')

    def _to_base_type(self, value):
        if value is None:
            return None
        return value.value

    def _from_base_type(self, value):
        if value is None:
            return None
        return next((item for item in self._enum_class if item.value == value), None)


class EncryptedProperty(ndb.BlobProperty):
    """Property that stores encrypted bytes.

    Encrypts bytes values using AES-256-GCM before storing in the datastore,
    and decrypts them when reading back.

    The AES-256-GCM key should be in the ``encrypted_property_key`` file, base64
    encoded. Here's example code to generate an AES-256-GCM key and base64 encode it:

      import base64
      from cryptography.hazmat.primitives.ciphers.aead import AESGCM

      key_bytes = AESGCM.generate_key(bit_length=256)
      print(base64.b64encode(key_bytes))
    """
    def _validate(self, value):
        if value is not None and not isinstance(value, bytes):
            raise TypeError('EncryptedProperty value must be bytes')

    def _to_base_type(self, value):
        if value is None:
            return None

        if not ENCRYPTED_PROPERTY_KEY:
            raise RuntimeError('No encryption key found in encrypted_property_key.pem')

        nonce = os.urandom(12)  # 96-bit nonce for GCM
        ciphertext = ENCRYPTED_PROPERTY_KEY.encrypt(nonce, value, None)

        # concatenate nonce and ciphertext for storage
        return nonce + ciphertext

    def _from_base_type(self, value):
        if value is None:
            return None

        if not ENCRYPTED_PROPERTY_KEY:
            raise RuntimeError('No encryption key found in encrypted_property_key.pem')

        nonce = value[:12]
        ciphertext = value[12:]
        return ENCRYPTED_PROPERTY_KEY.decrypt(nonce, ciphertext, None)


class Cache(ndb.Model):
    """Simple, dumb, datastore-backed key/value cache."""
    value = ndb.BlobProperty()
    expire = ndb.DateTimeProperty(tzinfo=timezone.utc)

    @classmethod
    def get(cls, key):
        """
        Args:
          key (str)

        Returns:
          str or None: value
        """
        if got := cls.get_by_id(key):
            if not got.expire or datetime.now(timezone.utc) < got.expire:
                return got.value.decode()

    @classmethod
    def put(cls, key, value, expire=None):
        """
        Args:
          key (str)
          value (str)
          expire (datetime.timedelta)
        """
        cached = cls(id=key, value=value.encode(),
                     expire=datetime.now(timezone.utc) + expire)
        super(cls, cached).put()


def stored_value(entity, prop):
  """Returns an entity's stored or computed value for a given property.

  This lets you the value that a entity currently has for a
  :class:`ndb.ComputedProperty` *without* actually computing it.

  This uses :class:`Model` 's undocumented internal ``_values`` attr. Ugh.

  Args:
    entity (ndb.Model)
    prop (str)
  """
  if val := entity._values.get(prop):
    return val.b_val if isinstance(val, _BaseValue) else val



"""Base OAuth flow views. Clients should use the individual site modules.

Example usage::

    app = Flask()
    app.add_url_rule('/start',
                     view_func=twitter.Start.as_view('start', '/callback'),
                     methods=['POST'])
    app.add_url_rule('/callback',
                     view_func=twitter.Callback.as_view('callback', '/after'))
"""
import logging
import urllib.parse

from flask import redirect, request, session
from flask.views import View
from google.cloud.ndb.key import Key

from .webutil import util

LOGINS_SESSION_KEY = 'oauth-dropins.logins'

logger = logging.getLogger(__name__)


class BaseView(View):
  """Base view class. Provides the to() factory method.

  Attributes:
    DEFAULT_SCOPE (str): default OAuth scope(s) to request
    SCOPE_SEPARATOR (str): used to separate multiple scopes
    LABEL (str): human-readable label, eg 'Bluesky'
    NAME (str): module name; usually same as `__name__.split('.')[-1]`
    to_path (str): the base redirect URL path for the OAuth callback
    scope (str): OAuth scopes, comma-separated
  """
  DEFAULT_SCOPE = ''
  SCOPE_SEPARATOR = ','
  LABEL = None
  NAME = None

  to_path = None
  scope = None

  def __init__(self, to_path, scopes=None):
    super().__init__()
    assert to_path
    self.to_path = to_path
    self.scope = self.make_scope_str(scopes)

  @classmethod
  def make_scope_str(cls, extra):
    """Returns an OAuth scopes query parameter value.

    Combines :attr:`DEFAULT_SCOPE` and extra.

    Args:
      extra (sequence of str, or None)
    """
    if not extra:
      return cls.DEFAULT_SCOPE

    if not isinstance(extra, str):
      extra = cls.SCOPE_SEPARATOR.join(extra)

    return cls.SCOPE_SEPARATOR.join(util.trim_nulls((cls.DEFAULT_SCOPE, extra)))

  def to_url(self, state=None):
    """Returns a fully qualified callback URL based on ``to_path``.

    Includes scheme, host, and optional state.
    """
    url = urllib.parse.urljoin(request.host_url, self.to_path)
    if state:
      # unquote first or state will be double-quoted
      state = urllib.parse.unquote_plus(state)
      url = util.add_query_params(url, [('state', state)])
    return url

  def request_url_with_state(self):
    """Returns the current request URL, with the state query param if provided."""
    state = request.values.get('state')
    if state:
      return util.add_query_params(request.base_url, [('state', state)])
    else:
      return request.base_url


class Start(BaseView):
  """Base class for starting an OAuth flow.

  Users should use the :meth:`to` class method when using this view in a WSGI
  application. See the file docstring for details.

  If the ``state`` query parameter is provided in the request data, it will be
  returned to the client in the OAuth callback view. If the ``scope`` query
  parameter is provided, it will be added to the existing OAuth scopes.

  Alternatively, clients may call :meth:`redirect_url` and HTTP 302 redirect to
  it manually, which will start the same OAuth flow.
  """

  def dispatch_request(self):
    scopes = set(request.values.getlist('scope'))
    if self.scope:
      scopes.add(self.scope)
    self.scope = self.SCOPE_SEPARATOR.join(util.trim_nulls(scopes))

    # str() is since WSGI middleware chokes on unicode redirect URLs :/ eg:
    # InvalidResponseError: header values must be str, got 'unicode' (u'...') for 'Location'
    # https://console.cloud.google.com/errors/CPafw-Gq18CrnwE
    url = str(self.redirect_url(state=request.values.get('state')))

    logger.info(f'Starting OAuth flow: redirecting to {url}')
    return redirect(url)

  def redirect_url(self, state=None):
    """Returns the local URL for the OAuth service to redirect back to.

    Subclasses must implement this.

    Args:
      state (str): user-provided value to be returned as a query parameter in
        the return redirect
    """
    raise NotImplementedError()

  @classmethod
  def button_html(cls, to_path, form_classes='', form_method='post',
                  form_extra='', image_prefix='', image_file=None,
                  input_style='', scopes='', outer_classes=''):
    """Returns an HTML string with a login form and button for this site.

    Args:
      to_path (str): path or URL for the form to POST to
      form_classes (str): optional, HTML classes to add to the <form>
      form_classes (str): optional, HTML classes to add to the outer <div>
      form_method (str): optional, form action ie HTTP method, eg 'get';
        defaults to 'post'
      form_extra (str): optional, extra HTML to insert inside the <form>
        before the button
      scopes (str): optional, OAuth scopes to override site's default(s)
      image_prefix (str): optional, prefix to add to the beginning of image
        URL path, eg '/oauth_dropins/'
      image_file (str): optional, image filename. defaults to [cls.NAME].png
      input_style (str): optional, inline style to apply to the button <input>

    Returns:
      str:
    """
    if image_file is None:
      image_file = f'{cls.NAME}_2x.png'
    vars = locals()
    vars.update({
      'label': cls.LABEL,
      'image': urllib.parse.urljoin(image_prefix, image_file),
    })
    html = f"""<form method="{vars['form_method']}" action="{vars['to_path']}" class="{vars['form_classes']}">
  <nobr>
    {vars['form_extra']}
    <input type="image" height="50" title="{vars['label']}" class="shadow"
           src="{vars['image']}" style="{vars['input_style']}" />
    <input name="scope" type="hidden" value="{vars['scopes']}">
  </nobr>
</form>
"""
    if outer_classes:
      html = f'<div class="{outer_classes}">{html}</div>'
    return html


class Callback(BaseView):
  """Base OAuth callback view.

  Users can use :meth:`to_url` when using this view in a WSGI application to
  make it redirect to a given URL path on completion. See file docstring for
  details.

  Alternatively, you can subclass it and implement :meth:`finish`, which will be
  called in the OAuth callback request directly, after the user has been
  authenticated.

  The auth entity and optional state parameter provided to Start will be
  passed to :meth:`finish` or as query parameters to the redirect URL.
  """

  def finish(self, auth_entity, state=None):
    """Called when the OAuth flow is complete. Clients may override.

    Args:
      auth_entity (models.BaseAuth): resulting auth entity, or None if the user
        declined the site's OAuth authorization request.
      state (str): passed to :meth:`Start.redirect_url`

    Returns:
      werkzeug.wrappers.Response:
    """
    if auth_entity is None:
      params = [('declined', True)]
    else:
      params = [('auth_entity', auth_entity.key.urlsafe().decode()),
                ('state', state)]
      try:
        token = auth_entity.access_token()
        if isinstance(token, str):
          params.append(('access_token', token))
        elif token:
          params += [('access_token_key', token[0]),
                     ('access_token_secret', token[1])]
      except NotImplementedError:
        logger.info('access_token() not implemented')

      try:
        token = auth_entity.refresh_token
        params.append(('refresh_token', token))
      except AttributeError:
        logger.info('refresh_token not included')

      login = (auth_entity.__class__.__name__, auth_entity.key.id())
      logins = session.setdefault(LOGINS_SESSION_KEY, [])
      if login not in logins:
        logins.append(login)
      session.modified = True
      session.permanent = True

    url = util.add_query_params(self.to_path, params)
    logger.info(f'Finishing OAuth flow: redirecting to {url}')
    return redirect(url)


def get_logins():
  """Returns all current logged in sessions, as auth entity keys.

  Returns:
    list of :class:`google.cloud.ndb.key.Key`: logged in auth entities
  """
  return [Key(kind, id) for kind, id in session.get(LOGINS_SESSION_KEY, [])]


def logout(auth=None):
  """Clears one login, or all, in the current Flask session.

  Args:
    auth (models.BaseAuth): login to remove from the session. Defaults to all.
  """
  if auth:
    key = (auth.__class__.__name__, auth.key.id())
    logins = session.get(LOGINS_SESSION_KEY, [])
    if key in logins:
      logins.remove(key)
  else:
    session.pop(LOGINS_SESSION_KEY, None)
